@@ -357,7 +357,7 @@ void centralized_done_with_process_id(int pid, int in_group)
         if (keep_unused) {
           st->next_unused = unused_pid_statuses;
           unused_pid_statuses = st;
-          if (st->signal_fd)
+          if (st->signal_fd && st->in_group)
             remove_group_signal_fd(st->signal_fd);
         } else
           st->unneeded = 1;
@@ -602,7 +602,6 @@ static void init_sigchld(rktio_t *rktio)
 #endif
 }
 
-#if !defined(CENTRALIZED_SIGCHILD)
 static void remove_from_sigchld_chain(rktio_t *rktio)
 {
   if (rktio->in_sigchld_chain) {
@@ -620,7 +619,6 @@ static void remove_from_sigchld_chain(rktio_t *rktio)
     }
   }
 }
-#endif
 
 static void check_child_done(rktio_t *rktio, pid_t pid)
 {
@@ -948,14 +946,25 @@ void rktio_process_forget(rktio_t *rktio, rktio_process_t *sp)
     centralized_ended_child();
   }
 # else
-  if (!((System_Child *)sp->handle)->done) {
+  System_Child *sc = (System_Child *)sp->handle, *sc2, *prev;
+  if (!sc->done) {
     void **unused_pid;
     unused_pid = malloc(sizeof(void *) * 2);
     unused_pid[0] = (void *)(intptr_t)sp->pid;
     unused_pid[1] = unused_pids;
     rktio->need_to_check_children = 1;
+
+    prev = NULL;
+    for (sc2 = rktio->system_children; sc2; prev = sc2, sc2 = sc2->next) {
+      if (sc2 == sc) {
+        if (prev)
+          prev->next = sc->next;
+        else
+          rktio->system_children = sc->next;
+      }
+    }
   }
-  free(sp->handle);
+  free(sc);
 # endif
 #endif
 
@@ -973,6 +982,10 @@ int rktio_process_init(rktio_t *rktio)
      it's a per-thread setting on Linux, and we want SIGCHLD blocked everywhere. */
   block_sigchld();
 
+  /* We'll set a handler later (possibly after other threads start),
+     so register SIGCHLD now: */
+  rktio_will_modify_os_signal_handler(SIGCHLD);
+
   centralized_start_child_signal_handler();
 #endif
 
@@ -1055,18 +1068,58 @@ static char *cmdline_protect(const char *s)
   return MSC_IZE(strdup)(s);
 }
 
+/* Avoid direct reference to Vista functionality: */
+typedef struct {
+  STARTUPINFOW StartupInfo;
+  void *lpAttributeList;
+} rktio_STARTUPINFOEXW;
+#define rktio_EXTENDED_STARTUPINFO_PRESENT 0x00080000
+#define rktio_PROC_THREAD_ATTRIBUTE_HANDLE_LIST 0x00020002
+typedef BOOL (*rktio_InitializeProcThreadAttributeList_t)(void *lpAttributeList,
+                                                          DWORD dwAttributeCount,
+                                                          DWORD dwFlags,
+                                                          PSIZE_T lpSize);
+static rktio_InitializeProcThreadAttributeList_t rktio_InitializeProcThreadAttributeList;
+typedef BOOL (*rktio_UpdateProcThreadAttribute_t)(void *lpAttributeList,
+                                                  DWORD dwFlags,
+                                                  DWORD_PTR Attribute,
+                                                  PVOID lpValue,
+                                                  SIZE_T cbSize,
+                                                  PVOID lpPreviousValue,
+                                                  PSIZE_T lpReturnSize);
+static rktio_UpdateProcThreadAttribute_t rktio_UpdateProcThreadAttribute;
+typedef void (*rktio_DeleteProcThreadAttributeList_t)(void *lpAttributeList);
+static rktio_DeleteProcThreadAttributeList_t rktio_DeleteProcThreadAttributeList;
+static void init_thread_attr_procs()
+{
+  if (!rktio_InitializeProcThreadAttributeList
+      || !rktio_UpdateProcThreadAttribute) {
+    HMODULE hm;
+
+    hm = LoadLibraryW(L"kernel32.dll");
+
+    rktio_InitializeProcThreadAttributeList = (rktio_InitializeProcThreadAttributeList_t)GetProcAddress(hm, "InitializeProcThreadAttributeList");
+    rktio_UpdateProcThreadAttribute = (rktio_UpdateProcThreadAttribute_t)GetProcAddress(hm, "UpdateProcThreadAttribute");
+    rktio_DeleteProcThreadAttributeList = (rktio_DeleteProcThreadAttributeList_t)GetProcAddress(hm, "DeleteProcThreadAttributeList");
+
+    FreeLibrary(hm);
+  }
+}
+
 static intptr_t do_spawnv(rktio_t *rktio,
                           const char *command, int argc, const char * const *argv,
 			  int exact_cmdline, intptr_t sin, intptr_t sout, intptr_t serr, int *pid,
 			  int new_process_group, int chain_termination_here_to_child,
-                          void *env, const char *wd)
+                          void *env, const char *wd,
+                          int disable_inherit)
 {
-  intptr_t i, l, len = 0;
+  intptr_t i, l, len = 0, retval;
   int use_jo;
   intptr_t cr_flag;
   char *cmdline;
   wchar_t *cmdline_w, *wd_w, *command_w;
-  STARTUPINFOW startup;
+  rktio_STARTUPINFOEXW startupx;
+  STARTUPINFOW *startup;
   PROCESS_INFORMATION info;
 
   if (exact_cmdline) {
@@ -1089,18 +1142,19 @@ static intptr_t do_spawnv(rktio_t *rktio,
     cmdline[len] = 0;
   }
 
-  memset(&startup, 0, sizeof(startup));
-  startup.cb = sizeof(startup);
-  startup.dwFlags = STARTF_USESTDHANDLES;
-  startup.hStdInput = (HANDLE)sin;
-  startup.hStdOutput = (HANDLE)sout;
-  startup.hStdError = (HANDLE)serr;
+  memset(&startupx, 0, sizeof(startupx));
+  startup = &startupx.StartupInfo;
+  startup->cb = sizeof(*startup);
+  startup->dwFlags = STARTF_USESTDHANDLES;
+  startup->hStdInput = (HANDLE)sin;
+  startup->hStdOutput = (HANDLE)sout;
+  startup->hStdError = (HANDLE)serr;
 
   /* If none of the stdio handles are consoles, specifically
      create the subprocess without a console: */
-  if (!rktio_system_fd_is_terminal(rktio, (intptr_t)startup.hStdInput)
-      && !rktio_system_fd_is_terminal(rktio, (intptr_t)startup.hStdOutput)
-      && !rktio_system_fd_is_terminal(rktio, (intptr_t)startup.hStdError))
+  if (!rktio_system_fd_is_terminal(rktio, (intptr_t)startup->hStdInput)
+      && !rktio_system_fd_is_terminal(rktio, (intptr_t)startup->hStdOutput)
+      && !rktio_system_fd_is_terminal(rktio, (intptr_t)startup->hStdError))
     cr_flag = CREATE_NO_WINDOW;
   else
     cr_flag = 0;
@@ -1133,25 +1187,61 @@ static intptr_t do_spawnv(rktio_t *rktio,
   wd_w = WIDE_PATH_copy(wd);
   command_w = WIDE_PATH_temp(command);
 
+  if (disable_inherit) {
+    /* don't just set the `bInherit` argument to `CreateProcessW` to
+       false, because that disables sharing for
+       stdin.stdout/stderr. */
+    init_thread_attr_procs();
+    if (rktio_InitializeProcThreadAttributeList
+        && rktio_UpdateProcThreadAttribute
+        && rktio_DeleteProcThreadAttributeList) {
+      LPPROC_THREAD_ATTRIBUTE_LIST lpAttributeList = NULL;
+      SIZE_T size = 0;
+      HANDLE handles_to_inherit[3];
+      rktio_InitializeProcThreadAttributeList(NULL, 1, 0, &size);
+      lpAttributeList = HeapAlloc(GetProcessHeap(), 0, size);
+      rktio_InitializeProcThreadAttributeList(lpAttributeList, 1, 0, &size);
+      handles_to_inherit[0] = startup->hStdInput;
+      handles_to_inherit[1] = startup->hStdOutput;
+      handles_to_inherit[2] = startup->hStdError;
+      rktio_UpdateProcThreadAttribute(lpAttributeList,
+                                      0, rktio_PROC_THREAD_ATTRIBUTE_HANDLE_LIST,
+                                      handles_to_inherit,
+                                      sizeof(handles_to_inherit), NULL, NULL);
+      startupx.lpAttributeList = lpAttributeList;
+      startup->cb = sizeof(startupx);
+      cr_flag |= rktio_EXTENDED_STARTUPINFO_PRESENT;
+    } else
+      disable_inherit = 0;
+  }
+
   if (cmdline_w
       && wd_w
       && command_w
       && CreateProcessW(command_w, cmdline_w, 
                         NULL, NULL, 1 /*inherit*/,
                         cr_flag, env, wd_w,
-                        &startup, &info)) {
+                        startup, &info)) {
     if (use_jo)
       AssignProcessToJobObject(rktio->process_job_object, info.hProcess);
     CloseHandle(info.hThread);
     *pid = info.dwProcessId;
     free(cmdline_w);
     free(wd_w);
-    return (intptr_t)info.hProcess;
+    retval = (intptr_t)info.hProcess;
   } else {
     if (cmdline_w) free(cmdline_w);
     if (wd_w) free(wd_w);
-    return -1;
+    retval = -1;
   }
+
+  if (disable_inherit) {
+    void *lpAttributeList = startupx.lpAttributeList;
+    rktio_DeleteProcThreadAttributeList(lpAttributeList);
+    HeapFree(GetProcessHeap(), 0, lpAttributeList);
+  }
+
+ return retval;
 }
 
 static void CopyFileHandleForSubprocess(intptr_t *hs, int pos)
@@ -1283,8 +1373,17 @@ rktio_process_result_t *rktio_process(rktio_t *rktio,
 
   if (envvars)
     env = rktio_envvars_to_block(rktio, envvars);
-  else
+  else {
+#ifdef RKTIO_USE_PTHREADS
+    {
+      rktio_envvars_t *current_envvars = rktio_envvars(rktio);
+      env = rktio_envvars_to_block(rktio, current_envvars);
+      rktio_envvars_free(rktio, current_envvars);
+    }
+#else
     env = NULL;
+#endif
+  }
 
 #if defined(RKTIO_SYSTEM_WINDOWS)
 
@@ -1320,7 +1419,8 @@ rktio_process_result_t *rktio_process(rktio_t *rktio,
 			     &pid,
                              new_process_group,
                              windows_chain_termination_to_child,
-                             env, current_directory);
+                             env, current_directory,
+                             flags & RKTIO_PROCESS_NO_INHERIT_FDS);
 
     if (!windows_exact_cmdline) {
       for (i = 0; i < argc; i++) {
@@ -1354,7 +1454,10 @@ rktio_process_result_t *rktio_process(rktio_t *rktio,
     init_sigchld(rktio);
 #endif
 
-    close_after_len = rktio_close_fds_len();
+    if (flags & RKTIO_PROCESS_NO_CLOSE_FDS)
+      close_after_len = 0;
+    else
+      close_after_len = rktio_close_fds_len();
 
     /* add a NULL terminator */
     {
@@ -1457,6 +1560,9 @@ rktio_process_result_t *rktio_process(rktio_t *rktio,
       if (env)
         free(env);
       free(new_argv);
+#if !defined(CENTRALIZED_SIGCHILD)
+      free(sc);
+#endif
 
       return NULL;
 
@@ -1514,6 +1620,8 @@ rktio_process_result_t *rktio_process(rktio_t *rktio,
         rktio_close_fds_after_fork(close_after_len, 0, 1, 2);
       }
 
+      rktio_restore_modified_signal_handlers();
+
       /* Set real CWD: */
       if (!rktio_set_current_directory(rktio, current_directory)) {
         fprintf(stderr, "racket: chdir failed to: %s\n", current_directory);
@@ -1524,11 +1632,14 @@ rktio_process_result_t *rktio_process(rktio_t *rktio,
 
       {
 	int err;
+        char **use_env;
 
         if (!env)
-          env = rktio_get_environ_array();
+          use_env = rktio_get_environ_array();
+        else
+          use_env = env;
 
-	err = MSC_IZE(execve)(command, (char **)new_argv, (char **)env);
+	err = MSC_IZE(execve)(command, (char **)new_argv, use_env);
         if (err)
           err = errno;
 
@@ -1547,6 +1658,8 @@ rktio_process_result_t *rktio_process(rktio_t *rktio,
 
     default: /* parent */
 
+      free(new_argv);
+
       break;
     }
 #endif
